/**
    Copyright (C) 1997 Pekka Nikander & Juha Prssinen.
    All rights Reserved.

    XXX   %Z%%M% version %I% date:%G%

    Module:  ArpTable (implementation)
    Package: arp
    File:    ArpTable.java


*/
package fi.hut.tcm.jacob.tcpip.arp;
import java.util.Date;
import fi.hut.tcm.jacob.tcpip.*;

/**
    ArpTable
    
    <br><br>
    The ARP table works as a cache memory of recently used IP addresses and
    their corresponding hardware addresses in
    order to prevent unnecessary ARP broadcasts for frequently used IP
    addresses. When the ARP protocol receives an
    outgoing data packet from a higher-level protocol, it consults the ARP
    table to see if the needed hardware address is
    already known.
    
*/
public class ArpTable {
    protected int[] ipAddrArr = new int[Arp.ArpTableSize];
    protected long[] etherAddrArr = new long[Arp.ArpTableSize];
    protected Date[] birthTimeArr = new Date[Arp.ArpTableSize];
    protected Date[] lastUsedArr = new Date[Arp.ArpTableSize];
    protected byte[] statusFlagArr = new byte[Arp.ArpTableSize];
    protected byte[] immutableFlagArr = new byte[Arp.ArpTableSize];
    protected boolean[] usedArr = new boolean[Arp.ArpTableSize];

    /**
        Inserts an entry into the ARP table. It is allowed to insert an
        entry for an IP address
        that already exists in the ARP table: the old entry will be
        overwritten.
        @param ipAddr   The IP address of the new entry.
        @param etherAddr   The ethernet address of the new entry.
        @param statusFlag   A status flag for the entry.
        @return   An indicator that shows if the insertion was successful.
    */
    public boolean insertEntry(int ipAddr, long etherAddr, byte statusFlag) {
    //+ ArpTable::insertEntry(int, long, byte)
      return(insertEntry(ipAddr, etherAddr, statusFlag, Arp.ARPMUTABLE));
    //- ArpTable::insertEntry(int, long, byte)
    }

    /**
        Inserts an entry into the ARP table. The immutable flag is
        optional. It is allowed to insert an entry for an IP address
        that already exists in the ARP table: the old entry will be
        overwritten.
        @param ipAddr   The IP address of the new entry.
        @param etherAddr   The ethernet address of the new entry.
        @param statusFlag   A status flag for the entry.
        @param ImmutableFlag  Permanent status yes/no
        @return   An indicator that shows if the insertion was successful.
    */
    public boolean insertEntry(int ipAddr, long etherAddr, byte statusFlag, byte ImmutableFlag) {
    //+ ArpTable::insertEntry(int, long, byte, byte)
      Date leastRecentlyUsed = new Date();

      int BestPlace = -1;
      int BestTime = -1;
      int k;

      for ( k = 0; k < Arp.ArpTableSize; k++ ) {   // See if the address is already there:
	if ( usedArr[k] ) {
	  if( ipAddr == ipAddrArr[k] ) {
	    if (immutableFlagArr[k] == 1) {
	      System.err.println("ArpTable: Error - tried to overwrite an immutable entry!");
	      return(false);
	    } else {
	      etherAddrArr[k] = etherAddr;
	      birthTimeArr[k] = new Date();
	      lastUsedArr[k] = new Date();
	      immutableFlagArr[k] = ImmutableFlag;
	      statusFlagArr[k] = statusFlag;
	      return(true);
	    }
	  }
	}
      }

    victim_search_loop:
      for( k = 0; k < Arp.ArpTableSize; k++) {   // Look for an empty or old slot

	if ( !usedArr[k] ) {
	  BestPlace = k;
	  BestTime = -1;
	  if (Arp.DebugLevel > 2) {  // #Debug#
	    System.out.print("Found empty slot in ArpTable");
	    System.out.println(" - slot " + k);
	  }
	  usedArr[k] = true;
	  break victim_search_loop;   // We're all done: we found an empty slot.
	} else if(lastUsedArr[k].getTime() < leastRecentlyUsed.getTime()) {
	  leastRecentlyUsed = lastUsedArr[k];
	  BestTime = k;
	  BestPlace = k;
	}
      }
      
      if (BestPlace != -1) {
	etherAddrArr[BestPlace] = etherAddr;
	ipAddrArr[BestPlace] = ipAddr;
	immutableFlagArr[BestPlace] = ImmutableFlag;
	birthTimeArr[BestPlace] = new Date();
	lastUsedArr[BestPlace] = new Date();
	statusFlagArr[BestPlace] = statusFlag;
      } else {
	System.out.println("ERROR: No room in ARP table!");
	return false; // No room in table!
      }
      
      return true;

    //- ArpTable::insertEntry(int, long, byte, byte)
    }

    /**
        Prints the ARP table, for debugging and direct user access. The
        following information is shown for each entry that is
        not too old: IP address, hardware address, status flag and
        immutable flag.
    */
    public void showTable() {
    //+ ArpTable::showTable( )

      System.out.println("==================== ArpTable.showTable ====================");
      for (int i = 0; i<Arp.ArpTableSize; i++) {
        System.out.println("Row: " + i + " [in use: " + usedArr[i] + "]");
	System.out.println("  IP: " + IpAddress.toString(ipAddrArr[i]) +
                           "  MAC: " + EthernetAddress.toString(etherAddrArr[i]));
	System.out.println("  Birth time: " + birthTimeArr[i]);
	System.out.println("  Last used:  " + lastUsedArr[i]);
	System.out.println("  Status flag = " + statusFlagArr[i] +
			   "  immutable flag = " + immutableFlagArr[i]);
      }
      System.out.println("------------------------------------------------------------");
    //- ArpTable::showTable( )
    }

    /**
        Deletes an entry from the ARP table using the IP address as the
        key. Returns true on success, and false if some error
        occurred.
        @param ipAddr   The IP address of the entry to be deleted
        @return   An indicator that shows if the delete was successful.
    */
    public boolean deleteEntry(int ipAddr) {
    //+ ArpTable::deleteEntry(int)

      for (int i = 0; i<Arp.ArpTableSize; i++ ) {
	if ( usedArr[i] && (ipAddr == ipAddrArr[i]) ) {
	  usedArr[i] = false;
	  return true;
	}
      }
      return false;  // Not found in table

    //- ArpTable::deleteEntry(int)
    }

    /**
        Returns the corresponding ethernet address in the given EtherAddr
        argument and the status for the requested IP
        address. If the IP address is not in the ARP table, or is too old,
        the returned value is NotInTable. Otherwise one of OK,
        or ARPFailure is returned. The contents of the EtherAddr argument
        are guaranteed to be valid only when the status
        code is OK. 
        @param ipAddr   The IP address of the desired Ethernet address.
        @return   The desired Ethernet address.
    */
    public long getEtherAddr(int ipAddr) {
    //+ ArpTable::getEtherAddr(int)

    boolean inTable = false;
    int j;
    
  find_entry_loop:
    for ( j = 0; j < Arp.ArpTableSize; j++) {
      if ( usedArr[j] && (ipAddr == ipAddrArr[j]) ) {
	inTable = true;
	break find_entry_loop;
      }
    }

    if( ! inTable ) 
      return(Arp.ARPNOTINTABLE);
    
    Date today = new Date();
    if ((statusFlagArr[j] == Arp.ARPFAILURE)
	&& (today.getTime() - Arp.ARPFAILEDTIME) < birthTimeArr[j].getTime()) {
      return(Arp.ARPFAILURE);
    } else {
      if (immutableFlagArr[j] == 1) {
	return etherAddrArr[j];
      } else {
	if((today.getTime() - Arp.ARPEXPIREDTIME) > birthTimeArr[j].getTime()) {
	  return(Arp.ARPNOTINTABLE);
	} else {
	  return etherAddrArr[j];
	  
	  // The following functionality is now missing:
	  //if((today.getTime() - Arp.ARPOKBUTOLDTIME) > BirthTime[j].getTime())
	  //  return(Arp.ARPOKBUTOLD);
	  //else return(Arp.ARPOK);
	}
      }
    }
    //- ArpTable::getEtherAddr(int)
    }

    public ArpTable() {
    //+ ArpTable::ArpTable( )
      for (int i = 0; i < Arp.ArpTableSize; i++) {
	usedArr[i] = false;
      }
    //- ArpTable::ArpTable( )
    }

}; //ArpTable

