/**
    Copyright (C) 1997 Pekka Nikander & Juha Prssinen.
    All rights Reserved.

    XXX   %Z%%M% version %I% date:%G%

    Module:  IcmpChecksumMessenger (implementation)
    Package: icmp
    File:    IcmpChecksumMessenger.java


*/
package fi.hut.tcm.jacob.tcpip.icmp;
import fi.hut.tcm.jacob.conduits.*;
import fi.hut.tcm.jacob.tcpip.*;

/**
    IcmpChecksumMessenger
    
    <p>An abstract class that provides the checksum algorithm for
    IcmpChecksumFromNetMessenger and IcmpChecksumToNetMessenger.
    The checksum algorithm calculates the ICMP checksum based on the packet
    header.
*/
public abstract class IcmpChecksumMessenger
  implements Messenger {

    /**
        Calculates the checksum.
                
        @param message    the <code>Message</code> for which to calculate
        the checksum
        @return    returns the checksum
    */
    protected int calculateChecksum(Message message) {
    //+ IcmpChecksumMessenger::calculateChecksum(Message)

      BinaryBuffer buf = message.getBuffer();
      int len = buf.getLength(); 
      //      byte[] data = buf.getBuffer();

      if (Icmp.DebugLevel > 1) {  // #Debug#
	System.out.println("Calculating ICMP checksum, packet dump:");
	buf.hexDump();
      }

      int elem;
      // optimization: one's complement of the checksum field,
      int csum = 0;
      // calculate the first one by hand:
      //elem = (((int) data[0] & 0xff) << 8) + ((int) data[1] & 0xff);
      elem = buf.getShort(0);
      csum += elem;
      // loop over the rest of the message:
      for (int i = 4; i < (len / 2) * 2; i += 2) {
	//elem = (((int) data[i] & 0xff) << 8) + (((int) data[i+1] & 0xff));
	elem = buf.getShort(i);
	csum += elem;
      }
      if ((len % 2) == 1) { // odd size
	//elem = (((int) data[len-1] & 0xff) << 8);
	elem = buf.getByte(len-1);
	csum += elem;
      }
      while ((csum >> 16) > 0) {
	csum = ((csum >> 16) & 0xffff) + (csum & 0xffff);
      }

      // final one's complement:
      csum = csum ^ 0xffff;

      if (Icmp.DebugLevel > 1) {  // #Debug#
	System.out.println("Calculated ICMP checksum = " + csum); // XXX Debug
      }

      return csum;
    //- IcmpChecksumMessenger::calculateChecksum(Message)
    }

    /**
        Gets the checksum for a <code>Message</code> containing one.
                
        @param message which <code>Message</code>
        
        @return    the checksum
    */
    protected int getChecksum(Message message) {
    //+ IcmpChecksumMessenger::getChecksum(Message)
      int csum;
      BinaryBuffer buf = message.getBuffer();
      csum = buf.getShort(2);
      if (Icmp.DebugLevel > 1) {  // #Debug#
	System.out.println("getChecksum: " + csum);
      }
      return csum;
    //- IcmpChecksumMessenger::getChecksum(Message)
    }

    /**
        Adds the ICMP checksum to a message.
                
        @param message  the <code>Message</code>
        @param csum    the checksum to add
    */
    protected void setChecksum(Message message, int csum) {
    //+ IcmpChecksumMessenger::setChecksum(Message, int)
      BinaryBuffer buf = message.getBuffer();

      buf.setShort(2, csum);

      if (Icmp.DebugLevel > 1) {  // #Debug#
	System.out.println("setChecksum: " + csum);
      }
    //- IcmpChecksumMessenger::setChecksum(Message, int)
    }

}; //IcmpChecksumMessenger

